/*******************************************************************
 *
 *	File:		CArloController.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/


#include <e32std.h>
#include <e32cons.h>
#include <f32file.h>
#include <e32svr.h>

#include "CArloController.h"
#include "CArlo.h"
#include "RArlo.h"
#include "CArloConfig.h"
#include "CConsoleEditor.h"

const TInt KArloMajor		= 1;
const TInt KArloMinor		= 22;
const TInt KArloMinorPatch	= 0;



#pragma warning (disable: 4238)

//#define DEBUG_PRINTS_PLEASE

#ifdef DEBUG_PRINTS_PLEASE
#define DBGPRINT(x) iConsole.Printf x
#define DBGPRESSKEY(x) iConsole.Printf x;iConsole.Getch()
#else
#define DBGPRINT(x) /**/
#define DBGPRESSKEY(x) /**/
#endif





CArloController::~CArloController()
{
	delete iArlo;
	delete iCommandLine;
	delete iArloConfig;
}


CArloController* CArloController::NewLC(CConsoleBase& aConsole, RArLo& aArloDevice, RFs& aFs)
{
	CArloController* self = new(ELeave)CArloController(aConsole, aFs);
	CleanupStack::PushL(self);
	self->ConstructL(aArloDevice);
	return self;
}


void CArloController::ConstructL(RArLo& aArloDevice)
{
	iArloConfig = CArloConfig::ReadConfigFileLC();
	CleanupStack::Pop();			// iArloConfig

	iArlo = ArloFactory::CreateArloLC(aArloDevice, iArloConfig->MachineName());
	CleanupStack::Pop();		// iArlo
	
//	iCommandLine = CLineEdit::NewL(&iConsole, 10);
	iCommandLine = CConsoleEditor::NewL(&iConsole);
}

void CArloController::DoTheArloThingL()
{
	DisplayBootMenuL(EFalse);		// normal boot menu

	TBool	finished = EFalse;
	TInt	timeout = iArloConfig->Timeout();
	while(!finished)
	{
		TChar	c;
		if (WaitForInput(c, timeout))
		{
			if ((c >= TChar('0')) && (c < (TChar('0' + iArloConfig->BootItems()))))
			{
				TInt item = c - TChar('0'); 
				const TArloBootItem&  bootItem = iArloConfig->BootItemL(item);
				BootLinuxL(bootItem);
				finished = ETrue;
			}
			else if (c == TChar('n'))
			{
				DisplayBootMenuL(EFalse);
			}
			else if (c == TChar('a'))
			{
				DisplayBootMenuL(ETrue);
			}
			else if (c == TChar('i'))
			{
				DisplayInformationL();
				DisplayBootMenuL(ETrue);
			}
			else if (c == TChar('e'))
			{
				iConsole.ClearScreen();
				
				if (!DoExpertArloThingL())
				{
					DisplayBootMenuL(ETrue);
				}
				else
				{
					finished = ETrue;
				}
			}
			else if (c == TChar('x'))
			{
				finished = ETrue;
			}
		}
		else
		{
			const TArloBootItem&  bootItem = iArloConfig->BootItemL(iArloConfig->DefaultBootItem());
			BootLinuxL(bootItem);
			finished = ETrue;
		}

		timeout = -1;
	}
}

void CArloController::DisplayBootMenuL(TBool aAdvanced)
{
	TMachineInfoV1Buf		machineInfo;

	iConsole.ClearScreen();

	if (aAdvanced)
	{
		TInt err = UserHal::MachineInfo(machineInfo);
		if (err == KErrNone)
		{
			if (iArlo->MachineName().CompareF(machineInfo().iMachineName) == 0)
			{
				iConsole.Printf(_L("Boot items from %S\n"), &iArloConfig->UsedConfigFile());
			}
			else
			{
				iConsole.Printf(_L("Boot items from %S, pretending to be a %S\n"), &iArloConfig->UsedConfigFile(), &iArlo->MachineName());
			}
		}
		else
		{
			iConsole.Printf(_L("Boot items from %S\n"), &iArloConfig->UsedConfigFile());
		}
	}
	else
	{
		TVersion	version(KArloMajor, KArloMinor, KArloMinorPatch);
		TMachineInfoV1Buf		machineInfo;

		iConsole.Printf(_L("ArLo for the Psion Series 5, 5mx and 7, version %S.\n"), &version.Name());

		TInt err = UserHal::MachineInfo(machineInfo);
		if (err != KErrNone)
			machineInfo().iMachineName.Format(_L("(unknown)"));
		
		if (machineInfo().iMachineName.CompareF(iArloConfig->MachineName()) == 0)
			iConsole.Printf(_L("Detected machine: %S\n"), &machineInfo().iMachineName);
		else
			iConsole.Printf(_L("Detected machine: %S, pretending to be a %S\n"), &machineInfo().iMachineName, &iArloConfig->MachineName());
		iConsole.Printf(_L("Author: Peter van Sebille (peter@yipton.demon.co.uk)\n\n"));

		iConsole.Printf(_L("Boot items:\n"));
	}

	if (iArloConfig->BootItems() > 0)
	{
		for (TInt i=0; i<iArloConfig->BootItems() ; i++)
		{
			if (iArloConfig->DefaultBootItem() == i)
				iConsole.Printf(_L(" *%c) %S\n"), TText('0') + i, &iArloConfig->BootItemL(i).iDescription);
			else
				iConsole.Printf(_L("  %c) %S\n"), TText('0') + i, &iArloConfig->BootItemL(i).iDescription);

			if (aAdvanced)
			{
				iConsole.Printf(_L("      kernel=%S\n"), &iArloConfig->BootItemL(i).iKernelFileName);
				iConsole.Printf(_L("      initrd=%S\n"), &iArloConfig->BootItemL(i).iInitrdFileName);
				iConsole.Printf(_L("      boot=%S\n"), &iArloConfig->BootItemL(i).iBootString);
			}
		}
	}
	else
	{
		iConsole.Printf(_L("     <no boot items>\n\n"));
	}
	
	if (aAdvanced)
	{
		iConsole.Printf(_L("  i) information\n"));
		iConsole.Printf(_L("  n) normal mode\n"));
		iConsole.Printf(_L("  e) expert mode\n"));
	}
	else
	{
		iConsole.Printf(_L("  a) advanced mode\n"));
	}
	iConsole.Printf(_L("  x) exit\n"));
}


TBool CArloController::WaitForInput(TChar& aChar, TInt aTimeout)
{
	TInt		x = iConsole.WhereX() + aTimeout;
	TInt		y = iConsole.WhereY();
	TInt		i;
	RTimer		timer;
	TBool		hasChar = EFalse;

	if (aTimeout == 0)
	{
		return hasChar;
	}

	for (i=0 ; i<aTimeout ; i++)
	{
		iConsole.Write(_L("."));
	}

	iConsole.SetCursorHeight(0);

	if (aTimeout > 0)
	{
		timer.CreateLocal();
		for (i=0 ; i<aTimeout ; i++)
		{
			TRequestStatus	stimer = KRequestPending;
			TRequestStatus	skey = KRequestPending;

			timer.After(stimer, 1000*1000*1);
			iConsole.Read(skey);
			User::WaitForAnyRequest();

			RDebug::Print(_L("WaitForInput second past"));

			if (skey != KRequestPending)
			{
				timer.Cancel();
				User::WaitForRequest(stimer);
				aChar = iConsole.KeyCode();
				iConsole.Printf(_L("%c\n"), aChar.GetLowerCase());
				hasChar = ETrue;
				break;
			}
			iConsole.ReadCancel();
			User::WaitForRequest(skey);

			iConsole.SetPos(--x, y);
			iConsole.Write(_L(" "));
			iConsole.Write(_L(" "));
			iConsole.SetPos(x, y);
		}
		timer.Close();
		iConsole.SetCursorHeight(100);
	}
	else
	{
		aChar = iConsole.Getch();
		RDebug::Print(_L("got key"));
		hasChar = ETrue;
	}

	return hasChar;
}

void CArloController::BootLinuxL(const TArloBootItem& aArloBootItem)
{
	iConsole.Printf(_L("\n\nBooting: %S\n"), &aArloBootItem.iDescription);
	
	iArlo->SetBootString((TAny *) aArloBootItem.iBootString.Ptr(), aArloBootItem.iDescription.Length());
	
	LoadFileL(aArloBootItem.iKernelFileName);
	if (aArloBootItem.iInitrdFileName.Length() > 0)
	{
		LoadFileL(aArloBootItem.iInitrdFileName);
	}

	iConsole.Printf(_L("\nBooting the kernel...\n"));
	iArlo->ArloBoot();

	// for WINS only...
	iConsole.Printf(_L("\nBye bye.... (press any key)\n"));
	iConsole.Getch();
}

void CArloController::DisplayInformationL()
{
	iConsole.ClearScreen();

	DisplayMachineInformationL();
	DisplayMemoryInformationL();
	DisplayVideoInformationL();

	iConsole.Printf(_L("\nPress any key to continue"));
	iConsole.Getch();
}

TBool CArloController::GetNumber(TLex& aLex, TUint32& aNumber)
{
	TChar		c1, c2;
	TBool		success = EFalse;

	if (aLex.Remainder().Length() > 2)
	{
		c1= aLex.Get();
		c2= aLex.Get();
		if (c1 == '0' && (c2 == 'x' || c2 == 'X'))
		{
			success = aLex.Val(aNumber, EHex) == KErrNone;
		}
		else
		{
			aLex.UnGet();
			aLex.UnGet();
		}
	}

	if (!success)
	{
		success = aLex.Val(aNumber, EDecimal) == KErrNone;
	}
	return success;
}

void CArloController::PrintExpertMenu()
{
	iConsole.ClearScreen();
	iConsole.Printf(_L("Commands are:   (any of these may crash/reset your machine)\n"));
	iConsole.Printf(_L("p <hex>      : convert virtual to physical address\n"));
	iConsole.Printf(_L("e <hex>      : examine word at virtual address, bypasses OS memory protection\n"));
	iConsole.Printf(_L("l <file>     : load file\n"));
	iConsole.Printf(_L("o <string>   : set boot string\n"));
	iConsole.Printf(_L("d <hex> <num>: dump num 32-bits words at virtual address\n"));
	iConsole.Printf(_L("w <hex>      : write 128 dwords 0xAAAAAAAA at virtual address\n"));
	iConsole.Printf(_L("c            : copy file (e.g., z:\\system\\libs\\ekern.exe) to c:\\ \n"));
	iConsole.Printf(_L("m            : print this menu\n"));
	iConsole.Printf(_L("b            : boot\n"));
#ifndef __WINS__
	iConsole.Printf(_L("r            : try to detect RAM layout\n"));
#endif
	iConsole.Printf(_L("a            : back to advanced mode\n"));
	iConsole.Printf(_L("x            : exit\n"));
}


void CArloController::DetectRam()
{
	TUint32		memoryDescription = 0x40001034;
	TUint32		ramStart = ArloExamine(memoryDescription);
	TUint32		ramLength;

	while (ramStart>0)
	{
		memoryDescription += 4;
		ramLength = ArloExamine(memoryDescription);
		if ((ramLength >> 20) > 0)
			iConsole.Printf(_L("  Ram bank: 0x%x, length = %dMB\n"), ramStart, ramLength >> 20);
		else
			iConsole.Printf(_L("  Ram bank: 0x%x, length = %dKB\n"), ramStart, ramLength >> 10);

		memoryDescription += 4;
		ramStart = ArloExamine(memoryDescription);
	}
}


TBool CArloController::DoExpertArloThingL()
{
	TBuf<80> cline;

 
	TUint32		number;
	TBool		finished = EFalse;
	TBool		wantExit = EFalse;

	PrintExpertMenu();

	do 
	{
		iCommandLine->EditLine(_L("> "), cline);

		TLex			lex(cline);
		TPtrC command = lex.NextToken();
		lex.SkipSpace();

		if (command == _L("p"))
		{
			TBool success = GetNumber(lex, number);
			if (!success || number == 0)
			{
				iConsole.Printf(_L("Illegal number\n"));
				continue;
			}

			ArloVirtualToPhysical(number);
		}
#ifndef __WINS__
		if (command == _L("r"))
		{
			DetectRam();
		}
#endif
		else if (command == _L("e"))
		{
			TBool success = GetNumber(lex, number);
			if (!success || number == 0)
			{
				iConsole.Printf(_L("Illegal number\n"));
				continue;
			}

			number = ArloExamine(number);
			iConsole.Printf(_L("ArloExamine: value = 0x%x\n"), number);
		}
		else if (command == _L("o"))
		{
			SetBootOptions(lex.NextToken());

		}
		else if (command == _L("d"))
		{
			TUint32		address;
			TUint32		count;

			TBool success = GetNumber(lex, address);
			if (!success || address == 0)
			{
				iConsole.Printf(_L("Illegal number\n"));
				continue;
			}
			lex.Assign(lex.NextToken());
			success = GetNumber(lex, count);
			if (!success || count == 0)
				count = 2;

			for (TUint i=0 ; i<count ; i++)
			{
				if (i % 4 == 0)
				{
					iConsole.Printf(_L("\n%08d: "), i / 4);
				}
				TUint	value = ArloExamine(address);
				iConsole.Printf(_L(" 0x%08x"), value);
				address += 4;
			}
			iConsole.Printf(_L("\n"));
		}
		else if (command == _L("w"))
		{
			TBool success = GetNumber(lex, number);
			if (!success || number == 0)
			{
				iConsole.Printf(_L("Illegal number\n"));
				continue;
			}

			TInt*  mem = (TInt*) number;
			for (int i=0 ; i<128 ; i++)
			{
				*mem++ = 0xAAAAAAAA;
			}
		}
		else if (command == _L("c"))
		{
			TPtrC	fileName = lex.NextToken();
		
			ArloCopyFile(fileName);
		}
		else if (command == _L("m"))
		{
			PrintExpertMenu();
		}
		else if (command == _L("l"))
		{
			TPtrC	fileName = lex.NextToken();
		
			LoadFile(fileName);
		}
		else if (command == _L("b"))
		{
			if (iArlo->HasLoadedFile())
			{
				iConsole.Printf(_L("\nBooting the kernel...\n"));
				iArlo->ArloBoot();
			}
			else
			{
				iConsole.Printf(_L("Failed, you have to load a file first\n"));
			}
		}
		else if (command == _L("a"))
		{
			finished = ETrue;
		}
		else if (command == _L("x"))
		{
			finished = ETrue;
			wantExit = ETrue;
		}
#if 0
		else if (command == _L("t"))
		{
			TUint32		number;
			TBool		success = GetNumber(lex, number);
			if (success)
				iConsole.Printf(_L("Success: number=0x%x\n"), number);
			else
				iConsole.Printf(_L("Failed\n"));
		}
#endif
		else if (command.Length() == 0)		// enter
		{
		}
		else
		{
			iConsole.Printf(_L("Unknown command: -%S-\n"), &command);
		}
	} while (!finished);

	return wantExit;
}


void CArloController::LoadFileL(const TDesC& aFileName)
{
	RFile		file;

	iConsole.Printf(_L("Opening file %S\n"), &aFileName);
	User::LeaveIfError(file.Open(iFs, aFileName, EFileRead));

	TInt size;
	User::LeaveIfError(file.Size(size));
	iConsole.Printf(_L("File size %x\n"), size);

	TUint8 *buf = new (ELeave) TUint8[size + PAGE_SIZE];
	TUint32 loadAdr = TUint32(buf);
	loadAdr += PAGE_SIZE - 1;
	loadAdr &= ~(PAGE_SIZE - 1);

	buf = (TUint8 *) loadAdr;
	TPtr fileBuf(buf, size);
	iConsole.Printf(_L("Loading %x bytes to address %08x\n"), size, loadAdr);
	User::LeaveIfError(file.Read(fileBuf));
	iConsole.Printf(_L("Read %x bytes\n"), fileBuf.Length());
	if (fileBuf.Length() != size)
	{
		User::Leave(KErrNotFound);
	}
	file.Close();

	TInt	blockNo = (size + PAGE_SIZE - 1) / PAGE_SIZE;    // No of 4k blocks needed
	iConsole.Printf(_L("No pages = %d\n"), blockNo);
	
	iArlo->AddPagesL(blockNo, loadAdr);
}

void CArloController::LoadFile(const TDesC& aFileName)
{
	TRAPD(err, LoadFileL(aFileName));
	if (err)
	{
		iConsole.Printf(_L("Error loading file err=0x%x\n"), err);
	}
}

void CArloController::DisplayMachineInformationL()
{
	TMachineInfoV1Buf		machineInfo;
	TBuf<64>				machineId;

	User::LeaveIfError(UserHal::MachineInfo(machineInfo));

	machineId.Num(machineInfo().iMachineUniqueId);
	iConsole.Printf(_L("Machine name  : %S, Machine Id: %S\n"), &machineInfo().iMachineName, &machineId);
	iConsole.Printf(_L("Processor     : %S (%d MHz), Speedfactor: %d, Processor Id: 0x%x\n"), &machineInfo().iProcessorName, (machineInfo().iProcessorClockInKHz / 1024), machineInfo().iSpeedFactor, iArlo->ArloGetProcId());
	iConsole.Printf(_L("Rom version   : %S\n"), &machineInfo().iRomVersion.Name());
	iConsole.Printf(_L("Keyboard name : %S\n"), &machineInfo().iKeyboardName);
	iConsole.Printf(_L("Display name  : %S, display size: %d x %d \n"), &machineInfo().iDisplayName, machineInfo().iDisplaySizeInPixels.iWidth, machineInfo().iDisplaySizeInPixels.iHeight);
	iConsole.Printf(_L("Input  name   : %S\n"), &machineInfo().iXYInputName);
	iConsole.Printf(_L("Psu name      : %S\n"), &machineInfo().iPsuName);
}

#define ONE_MB (1024 * 1024)

void CArloController::DisplayMemoryInformationL()
{
	TMemoryInfoV1Buf		memoryInfo;

	User::LeaveIfError(UserHal::MemoryInfo(memoryInfo));

	TBuf<8>	reprog = memoryInfo().iRomIsReprogrammable ? _L("is") : _L("is not");
	iConsole.Printf(_L("Memory        : %d MB Ram (%d MB Ramdisk), %d MB Rom, Rom %S reprogrammable\n"), memoryInfo().iTotalRamInBytes / ONE_MB, memoryInfo().iInternalDiskRamInBytes / ONE_MB, memoryInfo().iTotalRomInBytes / ONE_MB, &reprog);
}

void CArloController::DisplayVideoInformationL()
{
	TScreenInfoV01			screenInfo;
	TPckg<TScreenInfoV01>	sI(screenInfo);

	UserSvr::ScreenInfo(sI);

	iConsole.Printf(_L("Video memory  : linear address 0x%x\n"), screenInfo.iScreenAddress);
}


void CArloController::ArloVirtualToPhysical(TUint32 aVirtualAddress)
{
	TUint32		physicalAddress;
	
	iConsole.Printf(_L("ArloVirtualToPhysical: aVirtualAddress=0x%X\n"), aVirtualAddress);
	physicalAddress = iArlo->ArloLinear2Physical(aVirtualAddress);
	iConsole.Printf(_L("ArloVirtualToPhysical: physicalAddress=0x%X\n"), physicalAddress);
}


TUint CArloController::ArloExamine(TUint32 aAddressToExamine)
{
	TUint32		val = (TUint32) aAddressToExamine;

	DBGPRINT((_L("ArloExamine: peeking at 0x%x\n"), aAddressToExamine));
	DBGPRESSKEY((_L("Press a key\n")));

	val = iArlo->ArloExamine(val);

	DBGPRINT((_L("ArloExamine: value = 0x%x\n"), val));

	return (TUint) val;
}


void CArloController::CopyFileL(TDesC& aFileName)
{
	CFileMan*	fileman = CFileMan::NewL(iFs);
	User::LeaveIfError(fileman->Copy(aFileName, _L("c:\\"), ETrue));
	delete fileman;
}


void CArloController::ArloCopyFile(TDesC& aFileName)
{
	TRAPD(err, CopyFileL(aFileName));
	if (err)
	{
		iConsole.Printf(_L("Error copying file %S, err=%d\n"), &aFileName, err);
	}
	else
	{
		iConsole.Printf(_L("Successfully copied %S\n"), &aFileName);
	}
}

void CArloController::BootLinux(const TArloBootItem& aArloBootItem)
{
	TRAPD(err, BootLinuxL(aArloBootItem));
	iConsole.Printf(_L("Error in trying to boot, err=%d (press any key)\n"), err);
	iConsole.Getch();
}


void CArloController::SetBootOptions(const TDesC& aBootOptions)
{
	iArlo->SetBootString((TAny *) aBootOptions.Ptr(), aBootOptions.Length());
	iConsole.Printf(_L("Boot options set to: %S\n"), &aBootOptions);
}

